/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package Job_shop_transport;

import java.io.File;
import java.io.FileReader;
import java.io.StreamTokenizer;
import org.chocosolver.solver.Model;
import org.chocosolver.solver.Solution;
import org.chocosolver.solver.Solver;
import org.chocosolver.solver.variables.BoolVar;
import org.chocosolver.solver.variables.IntVar;
import org.chocosolver.util.ESat;

/**
 *
 * @author gondran
 */
public class Job_shop_transport {
    
    public static void main(String[] args) {
       
  
  /* -----------Declaration des variables ----------------------------*/
  int tableau_job_machine_veh[]={11,10,2};
   
  int Dmax = 670; // horizon de taille
  int[][] Machine = null;
  int[][] Processing_time = null;   
  int[] Taille_gamme = null;        // taille de la gamme
  int[][] Duree_deplacement=null;   // durée des déplacements
  
  int nb_job=tableau_job_machine_veh[0];
  int nb_machine=tableau_job_machine_veh[1];
  int nb_vehicule=tableau_job_machine_veh[2];
     
  Machine = new int[nb_job][nb_machine*2]; 
  Processing_time = new int[nb_job][nb_machine*2];   
  Duree_deplacement=new int[nb_machine+1][nb_machine+1];
  Taille_gamme=new int[nb_job];
  
  /* ----------- Lecture du fichier ----------------------------*/
  lire_fichier(Machine, Processing_time, Taille_gamme, Duree_deplacement, tableau_job_machine_veh);
  
  nb_job=tableau_job_machine_veh[0];
  nb_machine=tableau_job_machine_veh[1];
  nb_vehicule=tableau_job_machine_veh[2];

 
  affichage_gamme_et_processing_time( nb_job, Taille_gamme, Processing_time, Machine );

  affichage_traveling_time( nb_machine, Duree_deplacement);
  
  /* ------------Début ---------------------------*/
  
  Model mon_modele = new Model("JobShop avec transport");          
   
  // Contraintes 1,2,3 et 4  
  IntVar[][] St = new IntVar[nb_job][nb_machine*2+1]; 
  IntVar[][] Ft = new IntVar[nb_job][nb_machine*2+1];
  for (int i=0; i<nb_job; i++)
  {
    for (int j=0; j<Taille_gamme[i]; j++)
    {
        St[i][j] = mon_modele.intVar("St"+i+"_"+j, 0, Dmax);
        Ft[i][j] = mon_modele.intVar("Ft"+i+"_"+j, 0, Dmax);
    }
   }
     
  IntVar[][] v = new IntVar[nb_job][nb_machine*2+1]; 
  for (int i=0; i<nb_job; i++){
     for (int j=0; j<Taille_gamme[i]; j++){
         if (j%2==1){
            v[i][j] = mon_modele.intVar("v"+i+"_"+j, 1, nb_vehicule);
         }
         else{
            v[i][j] = mon_modele.intVar("v"+i+"_"+j, 0);
         }
     }
  }
  
  IntVar OBJ = mon_modele.intVar("Cmax", 0, Dmax);

   // durée de l'exécution d'une opération
  IntVar[][] durees = new IntVar[nb_job][nb_machine*2+1]; 
  for (int i=0; i<nb_job; i++){
     for (int j=0; j<Taille_gamme[i]; j++){
        if (j%2==0)// cas d'une opération machine
        {
          durees[i][j] = mon_modele.intVar("p_"+i+"_"+j, Processing_time[i][j]);
        }
        else // cas d'une opération de transport
        {
            int machine_depart=Machine[i][j-1];
            int machine_arrive=Machine[i][j+1];
            durees[i][j] = mon_modele.intVar("p_"+i+"_"+j,Duree_deplacement[machine_depart][machine_arrive]);
        }
     }
  }
  
     
  // bijonction des opérations machine
  BoolVar[][][][] b = new BoolVar [nb_job][nb_machine*2+1][nb_job][nb_machine*2+1];
  for (int i=0; i<nb_job; i++){
     for (int j=0; j<Taille_gamme[i]; j++){
         for (int u=0; u<nb_job; u++){
       for (int k=0; k<Taille_gamme[u]; k++){
        if (j%2==0 && k%2==0){
           b[i][j][u][k] =mon_modele.boolVar("b_"+i+"_"+j+"_"+u+"_"+k);  
        }
       }
    }
   }
  }  
  
  // dijonction vehicule
  IntVar[][][][] bv= new IntVar[nb_job][nb_machine*2+1][nb_job][nb_machine*2+1];
  for (int i=0; i<nb_job; i++){
     for (int j=0; j<Taille_gamme[i]; j++){
         for (int u=0; u<nb_job; u++){
           for (int k=0; k<Taille_gamme[u]; k++){
            if (!(i==u && j==k)){
             if (j%2!=0 && k%2!=0 ){
              bv[i][j][u][k] = mon_modele.intVar("bv_"+i+"_"+j+"_"+u+"_"+k,0,1);  
             }
            }
          }
       }
    }
  }  
  
  // contrainte 5. relation entre date de debut et de fin
  for (int i=0; i<nb_job; i++){
     for (int j=0; j<Taille_gamme[i]; j++){
        mon_modele.arithm(St[i][j],"+", durees[i][j],"=",Ft[i][j]).post();
     }
  }   

  // contrainte 6 : conjonction dans les gammes
  for (int i=0; i<nb_job; i++){
     for (int j=1; j<Taille_gamme[i]; j++){
        mon_modele.arithm(Ft[i][j-1], "<=",St[i][j]).post();
     }
  }     

  // contrainte 7 : disjonction entre les opérations machines
  for (int i=0; i<nb_job-1; i++){  
     for (int j=0; j<Taille_gamme[i]; j++){
        for (int u=i+1; u<nb_job; u++){
         for (int k=0; k<Taille_gamme[u]; k++){   
           if (j%2==0 && k%2==0){
            if (Machine[i][j]==Machine[u][k]){
              mon_modele.ifThenElse(b[i][j][u][k],
                                        mon_modele.arithm(Ft[i][j],"<=",St[u][k]),
                                        mon_modele.arithm(Ft[u][k],"<=",St[i][j]));
             }
           }  
         }
      }    
    }
  } 
  
  
  
    for (int i=0; i<nb_job-1; i++){  
        for (int j=0; j<Taille_gamme[i]; j++){
            for (int u=i+1; u<nb_job; u++){
                for (int k=0; k<Taille_gamme[u]; k++){   
                    if (!(i==u && j==k)){
                        if (j%2==1 && k%2==1){
                            mon_modele.ifThenElse(mon_modele.arithm(v[i][j],"=",v[u][k]),
                                                    mon_modele.arithm(bv[i][j][u][k],"+", bv[u][k][i][j],"=",1),
                                                    mon_modele.arithm(bv[i][j][u][k],"+", bv[u][k][i][j],"=",0)); 

                            int machine_ope_arrive_i_j=Machine[i][j+1];
                            int machine_ope_depart_u_v=Machine[u][k-1];
                            int transport_a_charge = Duree_deplacement[machine_ope_arrive_i_j][machine_ope_depart_u_v];
                            IntVar temps_transport=mon_modele.intVar("Duree_transport",transport_a_charge);
                            mon_modele.ifThen(mon_modele.arithm(bv[i][j][u][k],"=",1),
                                                mon_modele.arithm(temps_transport, "+", Ft[i][j],"<=",St[u][k]));
                    
                            int machine_ope_depart_i_j=Machine[i][j-1];
                            int machine_ope_arrive_u_v=Machine[u][k+1];
                            int transport_a_charge1 =Duree_deplacement[machine_ope_arrive_u_v][machine_ope_depart_i_j];
                            IntVar duree_depl=mon_modele.intVar("Duree_transport",transport_a_charge1);
                            mon_modele.ifThen(mon_modele.arithm(bv[u][k][i][j],"=",1),
                                                mon_modele.arithm(duree_depl, "+", Ft[u][k],"<=",St[i][j]));
                        }    
                    }
                }
            }    
        }
    }  

  
  // fonction objectif
  for (int i = 0; i < nb_job; i++){
    mon_modele.arithm(Ft[i][Taille_gamme[i]-1], "<=",OBJ).post();
  }
  Solver mon_solveur = mon_modele.getSolver();
  mon_modele.setObjective(Model.MINIMIZE, OBJ); 
    
  Solution s = new Solution(mon_modele);
  while (mon_solveur.solve()) {
     s.record();
  }  
       
  
  if (mon_solveur.isFeasible() == ESat.TRUE){
    affichage_affectation_vehicule( nb_job,  Taille_gamme, v,s);
    int cout =(int) mon_solveur.getBestSolutionValue(); // modif Matth 25/05
    System.out.println("Cmax = " + cout);
    affichage_dates_solution(nb_job, Taille_gamme,St, Ft, s);
   }
   else
     System.out.println("pas de sol.faisable");
 }
    
    
    
  // lecture des données du problème
    public static void lire_fichier(int[][] Machine,int[][] Processing_time,
                int[] Taille_gamme,int[][] Duree_deplacement,int[] tableau_job_machine_veh){
  
     int nb_job=11;   // jobs
     int nb_machine=10;  // machines
     int nb_vehicule=2;  // machines
  
     try{
  
     
        System.out.println(Job_shop_transport.class.getResource("Jobshop_T.class"));

         // lecture du jobset
         File  source;
         source = new File("./instance/JS1.txt");

         FileReader Input=new FileReader(source);
         StreamTokenizer   Lecteur;
         Lecteur=new StreamTokenizer (Input);
         Lecteur.parseNumbers();

         int nombre_ligne=-1;

         // nb jobs  
         Lecteur.nextToken();	
         if (Lecteur.ttype==Lecteur.TT_NUMBER){    
            nombre_ligne=(int) Lecteur.nval;
            nb_job=nombre_ligne;
         }

         // nb machines
         int nombre_colonne=-1;
         Lecteur.nextToken();
         if (Lecteur.ttype==Lecteur.TT_NUMBER){    
            nombre_colonne=(int) Lecteur.nval;
            nb_machine=nombre_colonne+1;
         }

         // nombre de véhicules
         Lecteur.nextToken();
         nb_vehicule=(int) Lecteur.nval;

        for (int i=0; i<nombre_ligne; i++){  
            Lecteur.nextToken();
            nombre_colonne=(int) Lecteur.nval; 
            Taille_gamme[i]=nombre_colonne*2+1; 
            Machine[i][0]=nb_machine; 

            for (int j=1; j<Taille_gamme[i]; j++){
                if(j%2==0) // opération machine
                {
                // lecture machine
                Lecteur.nextToken();
                if (Lecteur.ttype==Lecteur.TT_NUMBER){    
                    Machine[i][j]=(int) Lecteur.nval;
                }
                // lecture processing
                Lecteur.nextToken();
                if (Lecteur.ttype==Lecteur.TT_NUMBER){    
                Processing_time[i][j]=(int) Lecteur.nval;
                }
           }
           else{
            Machine[i][j]=0;
            Processing_time[i][j]=0;
           }
         }
        }

         // lecture des temps de déplacement
         source = new File("./instance/layout1.txt");
         Input=new FileReader(source);
         Lecteur=new StreamTokenizer (Input);
         Lecteur.parseNumbers();



          // dernière ligne
         Lecteur.nextToken();
         Duree_deplacement[nb_machine][nb_machine]= (int) Lecteur.nval; 
         for (int j=1; j<nb_machine; j++){ 
             Lecteur.nextToken();
             Duree_deplacement[nb_machine][j]=(int) Lecteur.nval;       
         }

          // autres lignes
        for (int i=1; i<nb_machine; i++){ 

           Lecteur.nextToken();
           Duree_deplacement[i][nb_machine]= (int) Lecteur.nval;     

            for (int j=1; j<nb_machine; j++){  
                Lecteur.nextToken();
                Duree_deplacement[i][j]=(int) Lecteur.nval;
            }
        }

         tableau_job_machine_veh[0]=nb_job;
         tableau_job_machine_veh[1]=nb_machine;
         tableau_job_machine_veh[2]=nb_vehicule;

     }catch (Exception e) {
        System.out.println(e.getMessage());
        System.out.println("erreur lors de la lecture");
    }      
 
  }
    
    
public static void affichage_affectation_vehicule(int nb_job, 
              int[] Taille_gamme,IntVar[][] vehicule_affecte, Solution s)
{
  System.out.println("Affectation des vehicules");
     
  for (int i=0; i<nb_job; i++){
    for(int j=0; j<Taille_gamme[i]; j++){
      if (j%2!=0 ) // on ne regarde que les opérations de transport
      {
        int camion = s.getIntVal(vehicule_affecte[i][j]);
        System.out.println(" --> Camion :  "+camion);
      }
    }
  }
}
    
    
    
// affiche des dates St et Ft de la solution
public static void affichage_dates_solution(int nb_job, int[] Taille_gamme,IntVar[][] St ,
          IntVar[][] Ft, Solution s)
{
  System.out.println("les dates = ");
  for (int i=0; i<nb_job; i++){
    for(int j=0; j<Taille_gamme[i]; j++) {
        int deb = s.getIntVal(St[i][j]);
        int fin = s.getIntVal(Ft[i][j]);
        System.out.print(deb + "/" + fin);
        System.out.print("    ");
    }
      System.out.println();
  }
}   
    
   
// affiche les durées de déplacement
public static void affichage_traveling_time(int nb_machine, int[][] Duree_deplacement)
{
  System.out.println();
  System.out.println();
  System.out.println("Traveling time");

  for (int i=0; i<=nb_machine; i++){
      for (int j=0; j<=nb_machine; j++){
         System.out.print(Duree_deplacement[i][j] + " ");
      }
      System.out.println();
  }
}
    
    
// affiche la gamme et les temps de process
public static void affichage_gamme_et_processing_time(int nb_job,int[] Taille_gamme,
        int[][] Processing_time,int[][] Machine )
{
  System.out.println("Gamme");
  for (int i=0; i<nb_job; i++){
     for (int j=0; j<Taille_gamme[i]; j++){
        System.out.print(Machine[i][j] + " ");
    }
    System.out.println();
  }

  System.out.println(); System.out.println();
  
  System.out.println("Processing time");
   
  for (int i=0; i<nb_job; i++){
    for (int j=0; j<Taille_gamme[i]; j++){
        System.out.print(Processing_time[i][j] + " ");
    }
    System.out.println();
  }
}



}
